@@ -1,3 +1,4 @@
+{-# OPTIONS_GHC -fno-warn-orphans #-}
 --  C->Haskell Compiler: monad for the binding generator
 --
 --  Author : Manuel M T Chakravarty
@@ -70,24 +71,28 @@
 module C2HS.Gen.Monad (
   TransFun, transTabToTransFun,
 
-  HsObject(..), GB, GBState(..), initialGBState, setContext, getLibrary, getPrefix,
+  HsObject(..), Wrapper(..), GB, GBState(..),
+  initialGBState, setContext, getLibrary, getPrefix,
   getReplacementPrefix, delayCode, getDelayedCode, ptrMapsTo, queryPtr,
-  objIs, queryObj, queryClass, queryPointer, mergeMaps, dumpMaps,
-  queryEnum, isEnum
+  objIs, queryObj, sizeIs, querySize, queryClass, queryPointer,
+  mergeMaps, dumpMaps, queryEnum, isEnum,
+  queryTypedef, isC2HSTypedef, queryDefaultMarsh, isDefaultMarsh,
+  addWrapper, getWrappers
 ) where
 
 -- standard libraries
-import Data.Char  (toUpper, toLower, isSpace)
+import Data.Char  (toUpper, toLower)
 import Data.List  (find)
 import Data.Maybe (fromMaybe)
 import qualified Data.Map as Map (empty, insert, lookup, union, toList, fromList)
 import Data.Map   (Map)
 import Data.Set   (Set)
-import qualified Data.Set as Set (empty, insert, member)
+import qualified Data.Set as Set (empty, insert, member, union, toList, fromList)
 
 -- Language.C
 import Language.C.Data.Position
 import Language.C.Data.Ident
+import Language.C.Syntax
 import Data.Errors
 
 -- C -> Haskell
@@ -95,7 +100,8 @@ import C2HS.C     (CT, readCT, transCT, raiseErrorCTExc)
 
 -- friends
 import C2HS.CHS   (CHSFrag(..), CHSHook(..), CHSTrans(..),
-                   CHSChangeCase(..), CHSPtrType(..))
+                   CHSChangeCase(..), CHSPtrType(..),
+                   CHSTypedefInfo, CHSDefaultMarsh, Direction(..))
 
 
 -- translation tables
@@ -104,7 +110,7 @@ import C2HS.CHS   (CHSFrag(..), CHSHook(..), CHSTrans(..),
 -- | takes an identifier to a lexeme including a potential mapping by a
 -- translation table
 --
-type TransFun = Ident -> String
+type TransFun = Ident -> Maybe String
 
 -- | translation function for the 'underscoreToCase' flag
 --
@@ -144,16 +150,18 @@ downcaseFirstLetter (c:cs) = toLower c : cs
 --   beginning of this file
 --
 transTabToTransFun :: String -> String -> CHSTrans -> TransFun
-transTabToTransFun prefx rprefx (CHSTrans _2Case chgCase table) =
-  \ide -> let
-            caseTrafo = (if _2Case then underscoreToCase else id) .
-                        (case chgCase of
-                           CHSSameCase -> id
-                           CHSUpCase   -> upcaseFirstLetter
-                           CHSDownCase -> downcaseFirstLetter)
-            lexeme = identToString ide
-            dft    = caseTrafo lexeme             -- default uses case trafo
-          in
+transTabToTransFun prefx rprefx (CHSTrans _2Case chgCase table omits) =
+  \ide ->
+  let caseTrafo = (if _2Case then underscoreToCase else id) .
+                  (case chgCase of
+                      CHSSameCase -> id
+                      CHSUpCase   -> upcaseFirstLetter
+                      CHSDownCase -> downcaseFirstLetter)
+      lexeme = identToString ide
+      dft    = caseTrafo lexeme             -- default uses case trafo
+  in if ide `elem` omits
+     then Nothing
+     else Just $
           case lookup ide table of                  -- lookup original ident
             Just ide' -> identToString ide'         -- original ident matches
             Nothing   ->
@@ -213,9 +221,34 @@ data HsObject    = Pointer {
                  deriving (Show, Read)
 type HsObjectMap = Map Ident HsObject
 
+type SizeMap = Map Ident Int
+
 -- | set of Haskell type names corresponding to C enums.
 type EnumSet = Set String
 
+-- Map from C type names to type default definitions.
+type TypedefMap = Map Ident CHSTypedefInfo
+
+-- Map from C type names to type default definitions.
+type DefaultMarshMap = Map (Direction, String, Bool) CHSDefaultMarsh
+
+-- Definitions for bare structure function wrappers.
+data Wrapper = Wrapper { wrapFn :: String
+                       , wrapOrigFn ::String
+                       , wrapDecl :: CDecl
+                       , wrapArgs :: [Bool]
+                       , wrapBools :: (Bool, [Bool])
+                       , wrapPos :: Position }
+             deriving Show
+
+instance Eq Wrapper where
+  w1 == w2 = wrapFn w1 == wrapFn w2
+
+instance Ord Wrapper where
+  compare w1 w2 = compare (wrapFn w1) (wrapFn w2)
+
+type WrapperSet = Set Wrapper
+
 {- FIXME: What a mess...
 instance Show HsObject where
   show (Pointer ptrType isNewtype) =
@@ -223,16 +256,35 @@ instance Show HsObject where
   show (Class   osuper  pointer  ) =
     "Class " ++ show ptrType ++ show isNewtype
 -}
+
+-- Remove everything until the next element in the list (given by a
+-- ","), the end of the list (marked by "]"), or the end of a record
+-- "}". Everything inside parenthesis is ignored.
+chopIdent :: String -> String
+chopIdent str = goChop 0 str
+    where goChop :: Int -> String -> String
+          goChop 0 rest@('}':_) = rest
+          goChop 0 rest@(',':_) = rest
+          goChop 0 rest@(']':_) = rest
+          goChop level ('(':rest) = goChop (level+1) rest
+          goChop level (')':rest) = goChop (level-1) rest
+          goChop level (_  :rest) = goChop level rest
+          goChop _     [] = []
+
+extractIdent :: String -> (Ident, String)
+extractIdent str =
+    let isQuote c = c == '\'' || c == '"'
+        (ideChars, rest) = span (not . isQuote)
+                           . tail
+                           . dropWhile (not . isQuote) $ str
+    in
+      if null ideChars
+      then error $ "Could not interpret " ++ show str ++ "as an Ident."
+      else (internalIdent ideChars, (chopIdent . tail) rest)
+
 -- super kludgy (depends on Show instance of Ident)
 instance Read Ident where
-  readsPrec _ ('`':lexeme) = let (ideChars, rest) = span (/= '\'') lexeme
-                             in
-                             if null ideChars
-                             then []
-                             else [(internalIdent ideChars, tail rest)]
-  readsPrec p (c:cs)
-    | isSpace c                                              = readsPrec p cs
-  readsPrec _ _                                              = []
+  readsPrec _ str = [extractIdent str]
 
 -- | the local state consists of
 --
@@ -254,7 +306,11 @@ data GBState  = GBState {
   frags     :: [(CHSHook, CHSFrag)], -- delayed code (with hooks)
   ptrmap    :: PointerMap,           -- pointer representation
   objmap    :: HsObjectMap,          -- generated Haskell objects
-  enums     :: EnumSet               -- enumeration hooks
+  szmap     :: SizeMap,              -- object sizes
+  enums     :: EnumSet,              -- enumeration hooks
+  tdmap     :: TypedefMap,           -- typedefs
+  dmmap     :: DefaultMarshMap,      -- user-defined default marshallers
+  wrappers  :: WrapperSet
   }
 
 type GB a = CT GBState a
@@ -267,7 +323,11 @@ initialGBState  = GBState {
                     frags  = [],
                     ptrmap = Map.empty,
                     objmap = Map.empty,
-                    enums = Set.empty
+                    szmap = Map.empty,
+                    enums = Set.empty,
+                    tdmap = Map.empty,
+                    dmmap = Map.empty,
+                    wrappers = Set.empty
                   }
 
 -- | set the dynamic library and library prefix
@@ -319,6 +379,10 @@ delayCode hook str  =
               && ide   == ide'   -> return frags'
             | otherwise          -> err (posOf ide) (posOf ide')
           Nothing                -> return $ frags' ++ [newEntry]
+      delay hook'@(CHSPointer _ _ _ _ _ _ _ _) frags' =
+        case find (\(hook'', _) -> hook'' == hook') frags' of
+          Just (CHSPointer _ _ _ _ _ _ _ _, _) -> return frags'
+          Nothing                              -> return $ frags' ++ [newEntry]
       delay _ _                                  =
         interr "GBMonad.delayCode: Illegal delay!"
       --
@@ -359,6 +423,21 @@ queryObj hsName  = do
                      fm <- readCT objmap
                      return $ Map.lookup hsName fm
 
+-- | add an entry to the size map
+--
+sizeIs :: Ident -> Int -> GB ()
+hsName `sizeIs` sz =
+  transCT (\state -> (state {
+                        szmap = Map.insert hsName sz (szmap state)
+                      }, ()))
+
+-- | query the size map
+--
+querySize       :: Ident -> GB (Maybe Int)
+querySize hsName  = do
+                     sm <- readCT szmap
+                     return $ Map.lookup hsName sm
+
 -- | query the Haskell object map for a class
 --
 -- * raise an error if the class cannot be found
@@ -401,14 +480,21 @@ mergeMaps     :: String -> GB ()
 mergeMaps str  =
   transCT (\state -> (state {
                         ptrmap = Map.union readPtrMap (ptrmap state),
-                        objmap = Map.union readObjMap (objmap state)
+                        objmap = Map.union readObjMap (objmap state),
+                        enums = Set.union readEnumSet (enums state)
                       }, ()))
   where
-    (ptrAssoc, objAssoc) = read str
+    -- Deal with variant interface file formats (old .chi files don't
+    -- contain the list of enumerations).
+    (ptrAssoc, objAssoc, enumList) =
+      case reads str of
+        [] -> let (ptr, obj) = read str in (ptr, obj, [])
+        [(r, "")] -> r
     readPtrMap           = Map.fromList [((isStar, internalIdent ide), repr)
                                         | ((isStar, ide), repr) <- ptrAssoc]
     readObjMap           = Map.fromList [(internalIdent ide, obj)
                                         | (ide, obj)            <- objAssoc]
+    readEnumSet          = Set.fromList enumList
 
 -- | convert the whole pointer and Haskell object maps into printable form
 --
@@ -416,10 +502,12 @@ dumpMaps :: GB String
 dumpMaps  = do
               ptrFM <- readCT ptrmap
               objFM <- readCT objmap
+              enumS <- readCT enums
               let dumpable = ([((isStar, identToString ide), repr)
                               | ((isStar, ide), repr) <- Map.toList ptrFM],
                               [(identToString ide, obj)
-                              | (ide, obj)            <- Map.toList objFM])
+                              | (ide, obj)            <- Map.toList objFM],
+                              Set.toList enumS)
               return $ show dumpable
 
 -- | query the enum map
@@ -435,6 +523,42 @@ isEnum :: String -> GB ()
 isEnum hsName =
   transCT (\state -> (state { enums = Set.insert hsName (enums state) }, ()))
 
+-- | query the type default map
+--
+queryTypedef :: Ident -> GB (Maybe CHSTypedefInfo)
+queryTypedef cIde  = do
+  tds <- readCT tdmap
+  return $ cIde `Map.lookup` tds
+
+-- | add an entry to the type default map
+--
+isC2HSTypedef :: Ident -> CHSTypedefInfo -> GB ()
+isC2HSTypedef cIde td =
+  transCT (\state -> (state { tdmap = Map.insert cIde td (tdmap state) }, ()))
+
+-- | query the default marshaller map
+--
+queryDefaultMarsh :: (Direction, String, Bool) -> GB (Maybe CHSDefaultMarsh)
+queryDefaultMarsh k  = do
+  dms <- readCT dmmap
+  return $ k `Map.lookup` dms
+
+-- | add an entry to the type default map
+--
+isDefaultMarsh :: (Direction, String, Bool) -> CHSDefaultMarsh -> GB ()
+isDefaultMarsh k dm =
+  transCT (\state -> (state { dmmap = Map.insert k dm (dmmap state) }, ()))
+
+-- | add a wrapper definition
+addWrapper :: String -> String -> CDecl ->
+              [Bool] -> (Bool, [Bool]) -> Position -> GB ()
+addWrapper wfn ofn cdecl args bools pos =
+  let w = Wrapper wfn ofn cdecl args bools pos
+  in transCT (\st -> (st { wrappers = Set.insert w (wrappers st) }, ()))
+
+getWrappers :: GB [Wrapper]
+getWrappers = Set.toList `fmap` readCT wrappers
+
 
 -- error messages
 -- --------------
@@ -467,4 +591,3 @@ hsObjExpectedErr ide  =
     ["Unknown name!",
      "`" ++ identToString ide ++ "' is unknown; it has *not* been defined by",
      "a previous hook."]
-
